! Pattern  - Seperate signal and noise with one stationary (t,x) PEF
!            Compute the signal using either the Soubaras/Wiener approach
!            or Nemeth approach
!
!
! USE: Pattern.x par= < in.H > ou.H 
!
! INPUTS:
!
!       in.H - float - data to be processed
!       Noise - char - sep file that contains the noise model
!       
!   PEF ESTIMATION:
!
!       nfiltern - integer - size of noise PEF -default [5,3]
!       nfilters - integer - size of signal PEF -default [5,3]
!       nitpefn  - integer - number of iterations for PEF estimation
!                            default 2*size(noise PEF) with CG method
!       nitpefs  - integer - number of iterations for PEF estimation
!                            default 2*size(signal PEF) with CG method
!       OPTIONAL:
!
!       Signal - char - sep file that contains a signal model, otherwise computes N*d
!
!       mask_noise -  char - sep file that contains the noise mask/weight
!                            default is mask_noise = 1.
!       mask_signal - char - sep file that contains the signal mask/weight
!                            default is mask_signal = 1.
!       PEF_noise_out - char - sep file that contains the coefficients
!                              of the noise PEF estimated
!       PEF_signal_out - char - sep file that contains the coefficients
!                              of the signal PEF estimated
!       PEF_noise_in - char - sep file that contains the coefficients
!                              of the noise PEF 
!       PEF_signal_in - char - sep file that contains the coefficients
!                              of the signal PEF
!   INVERSION:
!
!       niter - integer - number of iterations for the inversion default 50
!       remask - integer - 0 (default) if the datamask(weight) is not put to one before getting it from mask_sep
!                          1 otherwise.
!       nemeth - integer - 0 if Soubaras style (default)
!                          1 if Nemeth style
!       wantreg - integer - 0 Nemeth approach without regularization (default)
!                           1 Nemeth with damping (identity)
!       eps - float - regularization parameter (Soubaras/Nemeth) -default 1
!       mu - float - balancing parameter for Nemeth approach -default 1
!       verb - integer - verbosity level -default 1
!
!       OPTIONAL:
!
!       mask_sep - char - sep file for a signal/noise separation mask/weight
!
! OUTPUTS:
!
!         ou.H: panel 1 - estimated signal
!               panel 2 - inverse impulse response of the noise PEF
!                         should louk like the noise
!               panel 3 - inverse impulse response of the signal PEF
!
! TESTED IN 3D AND WORKS
!
!
!*Copyright 2002 Stanford University
!*All rights reserved
!
!*Author: Antoine Guitton

program Pattern 
 
  use sep
  use helix
  use bound
  use print
  use helicon
  use polydiv
  use pef_pat
  use polydiv2
  use compress
  use misinput
  use helicon2
  use cgstep_mod
  use solver2_mod
  use hell_row_mod
  use mat_row_init
  use createhelixmod

  implicit none

  real,   dimension (:), pointer             :: mn,ms
  real,   dimension (:), allocatable, target :: m
  real,   dimension (:), pointer             :: noise,signal,data,kk,kks,kkn
  real,   dimension (:), pointer             :: mask
  type (filter)                              :: Npef, Spef, Npeftmp, Speftmp
  
  integer, dimension(:), allocatable         :: nfiltn,n0n,n,gapn,centern,nin
  integer, dimension(:), allocatable         :: nfilts,n0s,gaps,centers
  integer, dimension(:), allocatable         :: nfiltern, nfilters
  integer                                    :: nitpefn, nitpefs, niter
  integer                                    :: ndim, n12, stat, wantreg
  real                                       :: eps, mu
  integer                                    :: verb, nemeth, i, j, remask

  call initpar()
  call sep_init     (SOURCE)
  ndim=sep_dimension()
  allocate(n(ndim-1))
  allocate(nin(ndim))
  call from_history('n',nin)
  n12=nin(1)*nin(2)
  n(1)=nin(1)
  n(2)=nin(2)

  allocate(noise(n12) , signal(n12), data(n12), kk(n12), &
             kks(n12),     kkn(n12), mask(n12))
  allocate(  m(2*n12))
  allocate(nfiltn(ndim-1),  nfilts(ndim-1))
  allocate(n0n   (ndim-1), centern(ndim-1), gapn(ndim-1), nfiltern(ndim-1))
  allocate(n0s   (ndim-1), centers(ndim-1), gaps(ndim-1), nfilters(ndim-1))
  
  mn => m(1:n12)
  ms => m(1+n12:)

  gapn    = 0
  n0n     = n
  centern = 1
  gaps    = 0


  n0s     = n
  centers = 1
  kk      = 1.
  kks     = 1.
  kkn     = 1.
  
  call from_param('niter'   ,niter, 50)
  call from_param('nfiltern',nfiltern,(/5,3/))
  call from_param('nfilters',nfilters, nfiltern)
  call from_param('wantreg' ,wantreg , 0)
  call from_param('eps'     ,eps     , 1.)
  call from_param('mu'      ,mu      , 1.)
  call from_param('verb'     ,verb    , 1)
  call from_param('nemeth'  ,nemeth  , 0)
  call from_param('remask'  ,remask  , 0)
  

  if (verb>0) write(0,*) '###############################'
  if (verb>0) write(0,*) '#    inversion parameters:    #'
  if (verb>0) write(0,*) '#'
  if (verb>0) write(0,*) '# Niter       =', niter
  if (verb>0) write(0,*) '# Pef Noise   =', nfiltern
  if (verb>0) write(0,*) '# Pef Signal  =', nfilters
  if (verb>0) write(0,*) '# Regularize? =', wantreg
  if (verb>0) write(0,*) '# eps (regul) =', eps
  if (verb>0) write(0,*) '# mu          =', mu
  if (verb>0) write(0,*) '# nemeth      =', nemeth
  if (verb>0) write(0,*) '# remask      =', remask

  where(nfiltern(2:)>1) centern (:ndim-1) = nfiltern(:ndim-1)/2 + 1
  where(nfilters(2:)>1) centers (:ndim-1) = nfilters(:ndim-1)/2 + 1
  
  Npef =  createhelix( n, centern, gapn, nfiltern)
  Spef =  createhelix( n, centers, gaps, nfilters)  
  Npeftmp =  createhelix( n, centern, gapn, nfiltern)
  Speftmp =  createhelix( n, centers, gaps, nfilters) 
  
  call boundn   ( n0n, n, nfiltern, Npef)
  call boundn   ( n0s, n, nfilters, Spef)
  call boundn   ( n0n, n, nfiltern, Npeftmp)
  call boundn   ( n0s, n, nfilters, Speftmp)
  
  call from_param('nitpefn ',nitpefn, size(Npef%lag)*2 )
  call from_param('nitpefs ',nitpefs, size(Spef%lag)*2 )
  
  if (verb>0) write(0,*) '# niter noise  pef = ',nitpefn
  if (verb>0) write(0,*) '# niter signal pef = ',nitpefs
  if (verb>0) write(0,*) '###############################'

   if (exist_file("mask_noise")) then
       write(0,*) '# Found mask for the noise model'
       call sep_read(kkn,"mask_noise")
    else
       kkn=1.
    end if
    
    if (exist_file("mask_signal")) then
       write(0,*) '# Found mask for the signal'
       call sep_read(kks,"mask_signal")
    else
       kks=1.
    end if
    
    if (j.gt.1) then
       Npef%mis   =0
       Spef%mis   =0
       Npeftmp%mis=0
       Speftmp%mis=0
    end if
     call find_mask((kkn /= 0.), Npef)
    call find_mask((kks /= 0.), Spef)  
    
    ! The remask option is useful when missing traces are present and we want
    ! to keep as many traces as possible.
    
    if (remask.eq.1) then
       
       mask=1.
       call find_mask((mask /= 0.), Npeftmp)
       call find_mask((mask /= 0.), Speftmp)
       
       if (exist_file("mask_sep")) then
          write(0,*) '# Found mask for the separation'
          call sep_read(mask,"mask_sep")
       end if
       
    else
       
       if (exist_file("mask_sep")) then
          write(0,*) '# Found mask for the separation'
          call sep_read(mask,"mask_sep")
       else
          mask=1.
       end if
       call find_mask((mask /= 0.), Npeftmp)
       call find_mask((mask /= 0.), Speftmp)
       
    end if
    where (Npeftmp%mis.or.Speftmp%mis) 
       kk = 0.
    elsewhere
       kk = mask
    end where
    
    write(0,*) '######################'
    write(0,*) '#                    #'
    write(0,*) '#     Noise  Pef     #'
    write(0,*) '#                    #'
    write(0,*) '######################'
    
    call sep_read (data)
    call sep_read (noise,'Noise')
     if (j.gt.1) then 
       write(0,*) 'Npef%lag',Npef%lag
     end if
    call find_pef_pat(noise, Npef, nitpefn, kkn)
    Npef%mis=Npeftmp%mis     
    call printn (n, centern, nfiltern, Npef)
    
    where(kk==0.)
       Npef%mis=.false.
    end where
    
    write(0,*) '#######################'
    write(0,*) '#                     #'
    write(0,*) '#     Signal  Pef     #'
    write(0,*) '#                     #'
    write(0,*) '#######################'
    
    if (exist_file("Signal")) then
       write(0,*) 'Found file Signal'
       call sep_read (signal,'Signal')
    else
       signal = 0.
       call helicon_init (Npef)
       stat = helicon_lop(.false.,.false.,data,signal)
    end if
   
    call find_pef_pat ( signal, Spef, nitpefs, kks)
    Spef%mis=Speftmp%mis
    call printn(n, centers, nfilters, Spef)
    
    where(kk==0.)
       Spef%mis=.false. 
    end where
  
    write(0,*) '#######################'
    write(0,*) '#                     #'
    write(0,*) '#   Begin inversion   #'
    write(0,*) '#                     #'
    write(0,*) '#######################'

    if (nemeth.eq.1) then
       call init_hell_row(n12,mu)
       call polydiv_init (n12,Spef)
       call polydiv2_init(n12,Npef)
    else
       call helicon_init(Npef)
       call helicon2_init(Spef,kk)
    end if
    
    do j=1,nin(3)

       call sep_read (data)

       if (nemeth.eq.1) then
          write(0,*) 'NEMETH STYLE WITH POLYDIV'
          
          if (wantreg.eq.0) then
             call solver(hell_row_op_smp, cgstep, niter=niter, x=m, &
             dat=data, wt=kk, verb=.true.)
          else
             call hell_mat_init(n12,n12)
             call solver_reg(hell_row_op_smp, cgstep, hell_mat_op, nreg=size(m), &
             x=m, dat=data, niter=niter, eps=eps, wt=kk, verb=.true.)
          end if
          
          where (kk.eq.0.)
             mask=0.
          elsewhere
             mask=1.
          end where
          stat = polydiv_lop(.false.,.false.,ms,signal)
          call sep_write(mu*mask*signal+data*(1-mask))
          !     call sep_write(kk)
          stat = polydiv2_lop(.false.,.false.,mn,signal)
          call sep_write(mask*signal+data*(1-mask))
        
          
       else
          
          write(0,*) 'SOUBARAS STYLE WITH HELICON'
        
          stat = helicon_lop(.false.,.false.,data,signal)
          call solver_reg(helicon_lop,cgstep,helicon2_lop,nreg=n12,x=m(1:n12),&
          dat=signal, niter=niter, eps=eps, wt=kk, verb=.true.)
          
          where (kk.eq.0.)
             mask=0.
          elsewhere
             mask=1.
          end where
          call sep_write(mask*m(1:n12)+data*(1-mask))
          

          
       end if
       
       
       
    end do
    
    call sep_close()
    
  end program Pattern





