!* Separ23d.x : Signal / Noise separation program with non-stationary t-x Prediction Error Filters.
!               Works in 2-D and 3-D (for better results)
!               See SEP2003, antoine[1-2-5-6] for examples
!
!  Separ23d.x < in.H par > out.H
!
!  Warning: needs Createmis.x in 3D for the boudary conditions of the PEFs.
!
!
! PARAMETERS:
!
!            in.H       = (nt,nx,ny) sep file that contains signal+noise.
!
!            out.H      = (nt,nx,ny)   if checkpef = 0 (see bellow)
!                         (nt,nx,3)    if checkpef = 1 and threeD=0
!                         (nt,nx,ny,3) if checkpef = 1 and threeD=1
!
!            mis3D.H    = (nt,nx,ny,2) Used only if threeD = 1, sep file containing the boundary conditions for
!                         the noise pef (first block) and the signal pef. These are computed with
!                         Createmis.x.
!
!            mask_sep   = (nt,nx,ny) [optional] sep file that contains a mask or a weight to be applied during the
!                       signal/noise separation
!
!            mask_noise = (nt,nx,ny) [optional] sep file that contains a mask or a weight to be applied during the
!                       PEF estimation for the noise (very useful if the noise has a lot of amplitude variation   
!                       I usually estimate a weight from an AGC
!            mask_signal= (nt,nx,ny) [optional] sep file that contains a mask or a weight to be applied during the
!                       PEF estimation for the signal (very useful if the noise has a lot of amplitude variation
!                       I usually estimate a weight from an AGC
!
!            Noise      = (nt,nx,ny) Noise model.
!
!            Signal     = (nt,nx,ny) [optional] Signal model. If a signal model is not provided, the signal PEF
!                         is estimated from N(d) directly, N being the noise PEF, and d the data.
!
!            Random     = (nt,nx,ny) [optional] File of random numbers to check the PEFs. To be used only with
!                         the nemeth style attenuation.
!
!***** PEF  ESTIMATION:   Minimize ||Ya|| + eps||a|| where  'a' is the PEF (unknown) and 'Y' the data
!
!        Noise -
!
!            na_noise   = size of the noise PEFs. Default [na_noise=5,3,3] [nt,nx,ny]
!
!         niter_noise   = number of iterations (CG) for estimating the noise PEFs
!                         Default (niter_noise=75)
!
!           eps_noise   = epsilon for the noise PEFs estimation, Defauls (eps_noise=1.)
!
!         npatch_noise  = number of micro-patches for the noise PEFs, Default (npatch_noise=10,10,1)
!
!
!         PEF_noise_out = [optional] output sep file that contains the PEF coefficients after estimation.
!         PEF_noise_in  = [optional] input sep file that contains the PEF coefficients.
!                          The PEFs are not estimated anymore.
!                          Warning: the PEFs should be consistent with na_ and npatch_
!        Signal -
!
!            na_signal  = size of the signal PEFs. Default [na_signal=5,3,3] [nt,nx,ny]
!
!         niter_signal  = number of iterations (CG) for estimating the signal PEFs
!                         Default (niter_signal=75)
!
!           eps_signal  = epsilon for the signal PEFs estimation, Defauls (eps_signal=1.)
!
!         npatch_signal = number of micro-patches for the signal PEFs, Default (npatch_signal=10,10,1)
!
!        PEF_signal_out = [optional] output sep file that contains the PEF coefficients after estimation.
!        PEF_signal_in  = [optional] input sep file that contains the PEF coefficients.
!                          The PEFs are not estimated anymore.
!                          Warning: the PEFs should be consistent with na_ and npatch_
!
!***** Signal/Noise separation
!
!      separmode = 0           minimizes ||Lsms+mu*Lnmn-d||+eps||ms,md|| (NEMETH STYLE)
!                  1 (default) minimizes ||N(s-d)||+eps||Ss|| (SOUBARAS STYLE)
!                  Soubaras is more stable than Nemeth, although Nemeth leads to better results.
!
!         threeD = 0 2-D noise attenuation (default)
!                  1 3-D noise attenuation
!
!       niter_sep = number of iterations (CG) for the signal/noise separation
!                   Default (niter_sep=75)
!
!         eps_sep = regularization operator for both Nemeth and Soubaras. Default (eps_sep=1.)
!
!          mu_sep = Balancing parameter for Nemeth (see above). Default (mu_sep=1.)
!
!          remask = 1 (default) PEFs are estimated according to the mask and the boundary conditions.
!                     The noise attenuation takes the boundary conditions into account only.
!                   0 PEFs are estimated according to the mask and the boundary conditions.
!                     The noise attenuation takes the boundary conditions and the masks into account.
!        checkpef = 0 (Default) Checks (or not) the PEFs.
!                   Convolve with the noise/"signal" model if separmode=1.
!                   Computes the impulse response          if separmode=0.
!                   The output goes in out.H
!
!            verb = verbosity level
!*Copyright 2002 Stanford University
!*All rights reserved
!
!*Author: Antoine Guitton
program sig_noise_separ
  
  use build_filter2d_mod
  use createnhelixmod
  use build_filter3d
  use nmisinput
  use npolydiv
  use nhelicon
  use npef_pat
  use nsignoi
  use random
  use sep 
  
  implicit none
  
  integer                         :: n(5)  
  real                            :: o(5),d(5) 
  real, dimension (:), pointer    :: data,noise_model,signal,kk
  real, dimension (:), allocatable:: tmpf,tmpf1,tmpf2  
  integer                         :: separmode, threeD             
  integer                         :: verb, remask, checkpef, nps, npn, nfs, nfn
  
  !SIGNAL FILTER PARAMETERS

  type(nfilter), pointer          :: signal_pef, signal_rough, signal_tmp, signal_rough_tmp      !PEF and roughener for the signal   
  integer                         :: na_signal(3),center_signal(3),gap_signal(3) !filter size
  integer                         :: niter_signal                                !number of iterations for PEF
  real                            :: eps_signal                                  !epsilon for regularization
  integer                         :: npatch_signal(3)                            !number of patches
  integer                         :: ncoef_signal
  real, dimension(:), pointer     :: kks, misnoise
  
  !NOISE FILTER PARAMETERS

  type(nfilter), pointer          :: noise_pef, noise_rough, noise_tmp, noise_rough_tmp        !PEF and roughener for the noise
  integer                         :: na_noise(3),center_noise(3),gap_noise(3) !filter size
  integer                         :: niter_noise                              !number of iterations for PEF
  real                            :: eps_noise                                !epsilon for regularization
  integer                         :: npatch_noise(3)                          !number of patches
  integer                         :: ncoef_noise
  real, dimension(:), pointer     :: kkn, missignal

  !SEPARATATION PARAMETERS

  integer                         :: niter_sep  !number of iterations
  real                            :: eps_sep    !epsilon for regularization
  real                            :: mu_sep     !Lnmn+mu*Lsms-d when signal and noise have
                                                !very different amplitudes

  !INTERNAL VARIABLES
  
  integer                         :: i3,n123,n45,ierr,stat,ndim,i
  integer,external                :: sreed
  
  call initpar()
 
  !read parameters
  call read_general_pars(verb,separmode,threeD,remask,checkpef,n,o,d,niter_sep,eps_sep,mu_sep)
  
  !signal filter
  call read_filter_pars("signal",na_signal,center_signal,gap_signal,niter_signal&
  &,eps_signal,npatch_signal)

  !noise filter
  call read_filter_pars("noise",na_noise,center_noise,gap_noise,niter_noise&
  &,eps_noise,npatch_noise)
  
  nps=product(npatch_signal) ! number of patches for the signal PEFs
  npn=product(npatch_noise)  ! number of patches for the noise PEFs

  !initialize parameters

  if (verb>0) write(0,*) ' '
  if (verb>0) write(0,*) ' '
  if (verb>0) write(0,*) ' '  
  if (verb>0) write(0,*) ' ------------------------------------'
  if (verb>0) write(0,*) ' '
  if (verb>0) write(0,*) '   2D-3D attenuation of noise with   '
  if (verb>0) write(0,*) '   non-stationary prediction-error   '
  if (verb>0) write(0,*) '               filters...            '
  if (verb>1) write(0,*) ' - - - - - - - - - - - - - - - - - - '
  if (verb>1) write(0,*) ' '
  if (verb>1) write(0,*) '        INVERSION PARAMETERS:        '
  if (verb>1) write(0,*) ' '                                    
  if (verb>1) write(0,*) ' ------------------------------------'
  if (verb>1) write(0,*) ' '
  if (verb>1) write(0,*) ' '
  if (verb>1) write(0,*) '      separmode   =  ',separmode
  if (verb>1) write(0,*) '         threeD   =  ',threeD
  if (verb>1) write(0,*) '         remask   =  ',remask
  if (verb>1) write(0,*) '       checkpef   =  ',checkpef
  if (verb>1) write(0,*) '      niter noise =  ',niter_noise  
  if (verb>1) write(0,*) '      niter signal=  ',niter_signal
  if (verb>1) write(0,*) '      niter separ =  ',niter_sep
  if (verb>1) write(0,*) ' '
  if (verb>1) write(0,*) '      eps noise   =  ',eps_noise
  if (verb>1) write(0,*) '      eps signal  =  ',eps_signal
  if (verb>1) write(0,*) '      eps separ   =  ',eps_sep
  if (verb>1) write(0,*) '      mu separ    =  ',mu_sep
  if (verb>1) write(0,*) ' '
  if (verb>1) write(0,*) '      PEF noise   =  ',na_noise
  if (verb>1) write(0,*) '      PEF signal  =  ',na_signal
  if (verb>1) write(0,*) '      patch noise =  ',npatch_noise
  if (verb>1) write(0,*) '      patch signal=  ',npatch_signal
  if (verb>1) write(0,*) ' '
  if (verb>0) write(0,*) ' ------------------------------------'
  if (verb>0) write(0,*) ' '

  allocate(signal_pef,signal_rough,signal_tmp,signal_rough_tmp)
  allocate(noise_pef,noise_rough,noise_tmp,noise_rough_tmp)
  
  if (threeD.eq.0) then

     center_signal(2)  = 1
     center_noise(2) = 1
     
     n123=n(1)*n(2)
     n45 =n(3)*n(4)*n(5)

     ! Build noise filters and rougheners
     if (verb>1) write(0,*) '  *** Build roughener/filter for 2D noise NS pefs ...'
     if (verb>1) write(0,*) ' '
     call init_build_filters2d(n(1:2),center_noise(1:2),na_noise(1:2),&
          & o(1:2),d(1:2),npatch_noise(1:2))
     call build_filters2d(noise_pef,noise_rough,ncoef_noise)
     call build_filters2d_smp(noise_tmp,ncoef_noise)

     ! Build signal filters and rougheners
     if (verb>1) write(0,*) '  *** Build roughener/filter for 2D signal NS pefs ...'
     if (verb>1) write(0,*) ' '
     call init_build_filters2d(n(1:2),center_signal(1:2),na_signal(1:2),&
          & o(1:2),d(1:2),npatch_signal(1:2))
     call build_filters2d(signal_pef,signal_rough,ncoef_signal)
     call build_filters2d_smp(signal_tmp,ncoef_signal)
  
  else

     n123=n(1)*n(2)*n(3)
     n45 =n(4)*n(5)

     allocate(misnoise(n123),missignal(n123))

     ! I do that because nbound in 3D takes FOR EVER!!!
     ! Therefore, I compute/load the boundary conditions
     ! before

     ! Load the mask for the misinput

     if (n123*4.ne. sreed("mis3D",misnoise,n123*4)) then
        call erexit("trouble reading mis noise")
     end if

     if (n123*4.ne. sreed("mis3D",missignal,n123*4)) then
        call erexit("trouble reading mis signal")
     end if

     ! Build noise filters and rougheners
     if (verb>1) write(0,*) '  *** Build roughener/filter for 3D noise NS pefs ...'
     if (verb>1) write(0,*) ' '
     call init_build_filters3d_smp(n,center_noise,na_noise,&
          & o,d,npatch_noise,misnoise)
     call build_filters3d2(noise_pef,noise_rough,ncoef_noise)
     call build_filters3d_smp2(noise_tmp,ncoef_noise)

     ! Build signal filters and rougheners
     if (verb>1) write(0,*) '  *** Build roughener/filter for 3D signal NS pefs ...'
     if (verb>1) write(0,*) ' '
     call init_build_filters3d_smp(n,center_signal,na_signal,&
          & o,d,npatch_signal,missignal)
     call build_filters3d2(signal_pef,signal_rough,ncoef_signal)
     call build_filters3d_smp2(signal_tmp,ncoef_signal)

     deallocate(missignal,misnoise)

  end if
 
  nfn=size(noise_pef%hlx(1)%lag)
  nfs=size(signal_pef%hlx(1)%lag)
  npn=size(noise_pef%hlx)
  nps=size(signal_pef%hlx)
  
  if (exist_file("PEF_noise_out")) then
     
     call to_history ("n1", nfn, "PEF_noise_out")
     call to_history ("n2", npn, "PEF_noise_out")
     call to_history ("n3", n45, "PEF_noise_out")
  
  end if
  
  if (exist_file("PEF_signal_out")) then
     
     call to_history ("n1", nfs, "PEF_signal_out")
     call to_history ("n2", nps, "PEF_signal_out")
     call to_history ("n3", n45, "PEF_signal_out")

  end if

  ! masks for the noise, signal  pef estimations and signal
  
  if (ierr.ne.0) then
     call erexit("trouble allocating")
  end if

  !loop over remaining axis

  do i3=1,n45 

     ! Allocation
     
     allocate(data(n123),noise_model(n123),&
     & signal(n123),kkn(n123),kks(n123),kk(n123),stat=ierr)
     
     if ((checkpef.eq.1).and.(separmode.eq.1)) then        
        allocate(tmpf(n123),tmpf1(n123),stat=ierr)
        if (ierr.ne.0) then
           call erexit("trouble allocating")
        end if        
     else if ((checkpef.eq.1).and.(separmode.eq.0)) then        
        allocate(tmpf(n123),tmpf1(n123),tmpf2(n123),stat=ierr)
        if (ierr.ne.0) then
           call erexit("trouble allocating")
        end if        
     end if
 
     kks=1.
     kkn=1.
     kk=1.
     
     !loop over data
     
     if (verb>0) write(0,*) "   *** Working on gather # ",i3,"/", n45
     if (verb>0) write(0,*) ' '
     data=0.

     if (n123*4.ne. sreed("in",data,n123*4)) then
        call erexit("trouble reading data")
     end if

     noise_model=0.
     if (n123*4.ne. sreed("Noise",noise_model,n123*4)) then
        call erexit("trouble reading noise")
     end if

     if (exist_file("mask_signal")) then
        if (n123*4.ne. sreed("mask_signal",kks,n123*4)) then
           call erexit("trouble reading mask_signal")
        end if
     end if
     
     if (exist_file("mask_noise")) then
        if (n123*4.ne. sreed("mask_noise",kkn,n123*4)) then
           call erexit("trouble reading mask_noise")
        end if
     end if
               
     call find_nmask(kks /=0.,signal_pef)
     call find_nmask(kkn /=0., noise_pef)

     if (remask.eq.1) then
        call find_nmask(kk /=0.,signal_tmp)
        call find_nmask(kk /=0., noise_tmp)
        
        if (exist_file("mask_sep")) then
           if (n123*4.ne. sreed("mask_sep",kk,n123*4)) then
              call erexit("trouble reading mask_sep")
           end if
        end if
    
     else

        if (exist_file("mask_sep")) then
           if (n123*4.ne. sreed("mask_sep",kk,n123*4)) then
              call erexit("trouble reading mask_sep")
           end if
        end if
        call find_nmask(kk /=0.,signal_tmp)
        call find_nmask(kk /=0.,noise_tmp)
      
     end if

     where(signal_tmp%mis.or.noise_tmp%mis)
        kk=0.
     end where

     call zero_nhelix(signal_pef)
     call zero_nhelix(noise_pef)
    
     if (exist_file("PEF_noise_in")) then
        
        if (verb>1) write(0,*) '     *** Read noise NS PEF coefficients ...'
        if (verb>1) write(0,*) ' '
        
        do i=1,npn
           if(nfn*4.ne.sreed("PEF_noise_in",noise_pef%hlx(i)%flt,nfn*4)) then
              call erexit("trouble reading PEF_noise_in")
           end if
        end do

     else

        if (verb>1) write(0,*) '     *** Estimate noise NS PEF ...'
        if (verb>1) write(0,*) ' '
        call find_pef(noise_model,noise_pef,noise_rough,niter_noise,&
        & eps_noise,ncoef_noise,kkn,(verb>2))

     end if

     if (exist_file("PEF_noise_out")) then
        
        if (verb>1) write(0,*) '     *** Write noise NS PEF coefficients ...'
        if (verb>1) write(0,*) ' '

        do i=1,npn
           call srite("PEF_noise_out",noise_pef%hlx(i)%flt,nfn*4)
        end do

     end if

     if (exist_file("PEF_signal_in")) then
       
        if (verb>1) write(0,*) '     *** Read signal NS PEF coefficients ...'
        if (verb>1) write(0,*) ' '
        
        do i=1,nps
           if(nfs*4.ne.sreed("PEF_signal_in",signal_pef%hlx(i)%flt,nfs*4)) then
              call erexit("trouble reading PEF_signal_in")
           end if
        end do

     else

        if (verb>1) write(0,*) ' '
        if (verb>1) write(0,*) '     *** Estimate signal NS PEF ...'
        if (verb>1) write(0,*) ' '
     
        if (exist_file("Signal")) then
           signal=0.
           if(n123*4.ne.sreed("Signal",signal,n123*4)) then
              call erexit("trouble reading signal")
           end if
        else
           signal=0.
           call nhelicon_init(noise_pef)
           stat=nhelicon_lop(.false.,.false.,data,signal)
           call nhelicon_close()
        end if
               
        call find_pef(signal,signal_pef,signal_rough,niter_signal,&
        & eps_signal,ncoef_signal,kks,(verb>2))
        
     end if
          
     if (exist_file("PEF_signal_out")) then
        
        if (verb>1) write(0,*) '     *** Write signal NS PEF coefficients ...'
        if (verb>1) write(0,*) ' '
        
        do i=1,nps
           call srite("PEF_signal_out",signal_pef%hlx(i)%flt,nfs*4)
        end do

     end if

     if ((checkpef.eq.1).and.(separmode.eq.1)) then
        
        call nhelicon_init(noise_pef)
        stat=nhelicon_lop(.false.,.false.,noise_model,tmpf)
        call nhelicon_close()
        
        where (noise_pef%mis)
           tmpf=0.
        end where

        if (maxval(signal)==0.) then
           signal=0.
           if (exist_file("Signal")) then
              if(n123*4.ne.sreed("Signal",signal,n123*4)) then
                 call erexit("trouble reading signal")
              end if
           else
              call nhelicon_init(noise_pef)
              stat=nhelicon_lop(.false.,.false.,data,signal)
              call nhelicon_close()
           end if
        end if

        call nhelicon_init(signal_pef)
        stat=nhelicon_lop(.false.,.false.,signal,tmpf1)
        call nhelicon_close()

        where (signal_pef%mis)
           tmpf1=0.
        end where

        call srite("out",tmpf, size(tmpf)*4)
        call srite("out",tmpf1,size(tmpf)*4)

     else if ((checkpef.eq.1).and.(separmode.eq.0)) then

        if (n123*4.ne.sreed("Random",tmpf2,n123*4)) then
           call erexit("trouble reading data")
        end if

        call npolydiv_init(n123,noise_pef)
        stat=npolydiv_lop(.false.,.false.,tmpf2,tmpf)
        call npolydiv_close()

        call npolydiv_init(n123,signal_pef)
        stat=npolydiv_lop(.false.,.false.,tmpf2,tmpf1)
        call npolydiv_close()
        call srite("out",tmpf, size(tmpf)*4)
        call srite("out",tmpf1,size(tmpf)*4)
        
     end if
     
     if ((checkpef.eq.1).and.(separmode.eq.1)) then
        deallocate(tmpf,tmpf1)
     else if ((checkpef.eq.1).and.(separmode.eq.0)) then
        deallocate(tmpf,tmpf1,tmpf2)
     end if
     deallocate(kks,kkn,noise_model)

     noise_pef%mis=noise_tmp%mis
     signal_pef%mis=signal_tmp%mis
     
     if (verb>1) write(0,*) ' '
     if (verb>1) write(0,*) '       *** Estimate the signal ...'
     if (verb>1) write(0,*) ' '
     
     call nsignoi_init(noise_pef,signal_pef,niter_sep,n123,eps_sep,&
     & separmode,mu_sep,(verb>2))     
     stat = nsignoi_lop( .false., .false., data, signal, kk)
     call  nsignoi_close()
    
     call srite("out",signal,size(signal)*4)
     
     deallocate(data,signal,kk)

  end do
  
  if (checkpef.eq.1) then   
     if (threeD.eq.0) then
        call to_history('n3',3*n(3)) 
     else
        call to_history('n4',3*n(4))
     end if
  end if
  
end program sig_noise_separ

subroutine read_general_pars(verb,separmode,threeD,remask,checkpef,n,o,d,niter_sep,eps_sep,mu_sep)
  
  use sep
 
  implicit none
  
  integer             :: separmode, threeD, remask, checkpef
  integer             :: verb, n(5)
  real                :: o(5), d(5)
  integer             :: niter_sep
  real                :: eps_sep, mu_sep
  integer             :: ndim,tempi,i
  character(len=1024) :: label
  real                :: tempr
  integer,external    :: getch,sep_get_number_data_axes,auxin

  !first DEAL with input data
  
  n(3)=1
  n(4)=1
  n(5)=1

  if (0.ne.sep_get_number_data_axes("in",ndim)) then
     call seperr("trouble obtaining data axes \n")
  end if

  if (ndim<2 .or. ndim>5) then
     call seperr("ndim must be between 2 and 5\n")
  end if
  
  do i=1,ndim
     call sep_get_data_axis_par("in",i,n(i),o(i),d(i),label)
  end do
  
  if (auxin("Noise").eq.-1) then
     call seperr("must provide noise file")
  end if
 
  if (0.ne.sep_get_number_data_axes("Noise",tempi)) then
     call seperr("trouble obtaining data axes for the noise file \n")
  end if
  
  if (tempi.ne.ndim) then
     call seperr("number of data axes and noise axes not equal\n")
  end if
  
  do i=1,ndim 
     call sep_get_data_axis_par("Noise",i,tempi,tempr,tempr,label)
     if (tempi.ne.n(i)) then
        call seperr("size of data axis ", i ," not equal noise axis\n")
     end if
  end do
  
  call from_param("verb",     verb,        0 ) 
  call from_param("niter_sep",niter_sep,   75)
  call from_param("eps_sep",  eps_sep,     1.)
  call from_param("mu_sep",   mu_sep,      1.)
  call from_param("separmode",separmode,   0 )
  call from_param("remask",   remask,      1 )
  call from_param("threeD",   threeD,      0 )
  call from_param("checkpef", checkpef,    0 )

end subroutine read_general_pars

subroutine read_filter_pars(string,na,center,gap,niter,eps,npatch)
  
  use sep
  
  implicit none
  
  character            (len=*) string
  integer              :: na(3),center(3),gap(3),niter,npatch(3)
  real                 :: eps
  character (len=1024) ::  temp
  integer,external     :: getch
  
  temp="na_"//string
  call from_param(temp,na,(/5,3,3/))
  
  temp="gap_"//string
  call from_param(temp,gap,(/0,0,0/))
  
  temp="center_"//string
  call from_param(temp,center,(/floor(real(na(1))/2.+.5),floor(real(na(2))/2+.5),1/))

  temp="npatch_"//string
  call from_param(temp,npatch,(/10,10,1/))

  temp="niter_"//string
  call from_param(temp,niter,75)
  
  temp="eps_"//string
  call from_param(temp,eps,1.)

end subroutine read_filter_pars

subroutine zero_nhelix(npef)
  
  use createnhelixmod

  type(nfilter)   :: npef
  integer         :: i

  do i=1,size(npef%hlx)
     npef%hlx(i)%flt=0.
  end do

end subroutine zero_nhelix
